"""
Read more about seaborn color palette
http://stanford.edu/~mwaskom/software/seaborn-dev/tutorial/color_palettes.html
"""
from ._xkcd_rgb import xkcd_rgb
from ._circos import circos

from ._palettes import generate_colors_palette

